#!/bin/bash

#*************************************************************************
#*************************************************************************
# Docker-compose backup script; see README.md for details/instructions
#*************************************************************************
#*************************************************************************

##############
# Constants
##############

COMPOSE_FILE_NAME="docker-compose.yml"

##############
# Config file defaults
##############

BACKUP_HOME=/opt/docker/backups
CONFIG_FILE=/opt/docker/.backupconfig
LOG_FILE_NAME=backups.log
PASSWORD=
PROJECTS_HOME=/opt/docker/projects
REMOTE_KEY=
REMOTE_SECRET=
REMOTE_URL=

##############
# Command line parameter initialization
##############

while getopts "b:c:h:k:l:p:s:u:" opt; do
	case ${opt} in
	b)
		BACKUP_HOME=$OPTARG
		;;
	c)
		CONFIG_FILE=$OPTARG
		;;
	h)
		PROJECTS_HOME=$OPTARG
		;;
	k)
		REMOTE_KEY=$OPTARG
		;;
	l)
		LOG_FILE_NAME=$OPTARG
		;;
	p)
		PASSWORD=$OPTARG
		;;
	s)
		REMOTE_SECRET=$OPTARG
		;;
	u)
		REMOTE_URL=$OPTARG
		;;
	*) ;;
	esac
done
shift $((OPTIND - 1))

##############
# Read config file
##############

if [ -f "$CONFIG_FILE" ]; then
	while IFS="=" read -r key value; do
		case "$key" in
		'#'*) ;;
		*)
			if [ -n "$key" ]; then
				eval "${key}=\"${value}\""
			fi
			;;
		esac
	done <"$CONFIG_FILE"
fi

##############
# Functions
##############

function with_slash() {
	STR=$1
	[[ "${STR}" != */ ]] && STR="${STR}/"
	echo "$STR"
}

function log() {
	echo "$(date +%FT%T.%3NZ): $1" | tee -a "$logFile"
}

function get_backend() {
	if [ -z "$REMOTE_KEY" ] || [ -z "$REMOTE_SECRET" ]; then
		echo LOCAL
	elif [[ "$REMOTE_URL" == s3* ]]; then
		echo AWS
	elif [[ "$REMOTE_URL" == azure* ]]; then
		echo AZURE
	elif [[ "$REMOTE_URL" == b2* ]]; then
		echo B2
	else
		echo LOCAL
	fi
}

function get_remote_key_name() {
	backend="$(get_backend)"
	if [ "$backend" == "AWS" ]; then
		echo "AWS_ACCESS_KEY"
	elif [ "$backend" == "AZURE" ]; then
		echo "AZURE_ACCOUNT_NAME"
	elif [ "$backend" == "B2" ]; then
		echo "B2_ACCOUNT_ID"
	fi
}

function get_remote_secret_name() {
	backend="$(get_backend)"
	if [ "$backend" == "AWS" ]; then
		echo "AWS_SECRET_ACCESS_KEY"
	elif [ "$backend" == "AZURE" ]; then
		echo "AZURE_ACCOUNT_KEY"
	elif [ "$backend" == "B2" ]; then
		echo "B2_ACCOUNT_KEY"
	fi
}

function get_remote_path() {
	repoName=$1
	backend="$(get_backend)"
	if [[ "$backend" == "AWS" ]]; then
		echo "$REMOTE_URL/$repoName"
	elif [[ "$backend" == "AZURE"  ]]; then
		echo "$REMOTE_URL:$repoName"
	elif [[ "$backend" == "B2" ]]; then
		echo "$REMOTE_URL:$repoName"
	fi
}

##############
# Exit if no password set
##############

if [ -z "$PASSWORD" ]; then
	log "No password set; exiting!"
	exit 1
fi

#############
# Script execution
##############

mkdir -p "$BACKUP_HOME"

if [ ! -d "$BACKUP_HOME" ]; then

	echo "$BACKUP_HOME could not be created"
	exit 1

fi

backend="$(get_backend)"

cachePath="$(with_slash "$BACKUP_HOME")"cache
mkdir -p "$cachePath"

logFile="$(with_slash "$BACKUP_HOME")$LOG_FILE_NAME"

errors=false

if [[ "$backend" == "LOCAL" ]]; then
	reposHomePath="$(with_slash $BACKUP_HOME)"repos
	mkdir -p "$reposHomePath"
else
	reposHomePath="$(with_slash $REMOTE_URL)"repos
fi

log "**** Starting backups ****"
log "**** Config file: $CONFIG_FILE"
log "**** Log file: $logFile"
log "**** Projects: $PROJECTS_HOME"
log "**** Repos: $reposHomePath"
log "**** Cache: $cachePath"
log "**** Backend: $(get_backend)"

while IFS= read -r -d '' file; do

	folder=${file%/$COMPOSE_FILE_NAME}
	repoName="${folder#$PROJECTS_HOME/}"
	repoName="${repoName//\//__}"
	backupPath="$reposHomePath"/"$repoName"
	remotePath="$(get_remote_path $repoName)"
	remoteKeyName="$(get_remote_key_name)"
	remoteSecretName="$(get_remote_secret_name)"
	volumeIds=()
	parameters=

	cd "$folder" || exit

	if [ -z "$(docker-compose ps -q)" ]; then
		# if [ "$folder" != "/opt/docker/projects/domains/nettek.at/wp-nettek" ]; then # for testing purposes

		log "Skipped file $folder: project is down"

	else

		containerIds="$(docker-compose ps -q)"
		for containerId in $containerIds; do
			for volumeId in $(docker inspect --format="{{ range .Mounts }}{{ .Name }} {{ end }}" "$containerId"); do
				if [ "$(docker volume inspect --format="{{ len .Labels }}" "$volumeId")" != 0 ]; then
					volumeIds+=("$volumeId")
				fi
			done
		done

		if [ ${#volumeIds[@]} = 0 ]; then

			log "Skipped $folder: no named volumes mounted"

		else

			parameters=" -v $folder:/data/configuration"
			for volumeId in "${volumeIds[@]}"; do
				parameters=" $parameters -v $volumeId:/data/volumes/$volumeId:ro"
			done

			if [ "$backend" == "LOCAL" ]; then
				
				docker run --rm \
					--name restic \
					--hostname "$(hostname)" \
					-e XDG_CACHE_HOME=/cache
					-e RESTIC_PASSWORD="$PASSWORD" \
					-v "$cachePath":/cache \
					-v "$backupPath":/repo \
					restic/restic snapshots -r /repo

				exitCode=$?

				if [ $exitCode != 0 ]; then

					docker run --rm \
						--name restic \
						--hostname "$(hostname)" \
						-e XDG_CACHE_HOME=/cache \
						-e RESTIC_PASSWORD="$PASSWORD" \
						-v "$cachePath":/cache \
						-v "$backupPath":/repo \
						restic/restic init -r /repo
			
				fi
			else

				docker run --rm \
					--name restic \
					--hostname "$(hostname)" \
					-e XDG_CACHE_HOME=/cache \
					-e RESTIC_PASSWORD="$PASSWORD" \
					-e "$remoteKeyName"="$REMOTE_KEY" \
					-e "$remoteSecretName"="$REMOTE_SECRET" \
					-v "$cachePath":/cache \
					restic/restic snapshots -r "$remotePath"

				exitCode=$?

				if [ $exitCode != 0 ]; then

					docker run --rm \
						--name restic \
						--hostname "$(hostname)" \
						-e XDG_CACHE_HOME=/cache \
						-e RESTIC_PASSWORD="$PASSWORD" \
						-e "$remoteKeyName"="$REMOTE_KEY" \
						-e "$remoteSecretName"="$REMOTE_SECRET" \
						-v "$cachePath":/cache \
						-e RESTIC_PASSWORD="$PASSWORD" \
						restic/restic init -r "$remotePath"
			
				fi
			fi

			docker-compose down

			if [ "$backend" == "LOCAL" ]; then

				docker run --rm \
					--name restic \
					--hostname "$(hostname)" \
					-e XDG_CACHE_HOME=/cache \
					-e RESTIC_PASSWORD="$PASSWORD" \
					-v "$cachePath":/cache \
					-v "$backupPath":/repo \
					$parameters \
					restic/restic -r /repo backup /data

				exitCode=$?
			
			else

				docker run --rm \
					--name restic \
					--hostname "$(hostname)" \
					-e XDG_CACHE_HOME=/cache \
					-e RESTIC_PASSWORD="$PASSWORD" \
					-e "$remoteKeyName"="$REMOTE_KEY" \
					-e "$remoteSecretName"="$REMOTE_SECRET" \
					-v "$cachePath":/cache \
					$parameters \
					restic/restic -r "$remotePath" backup /data

				exitCode=$?

			fi


			docker-compose up -d

			if [ "$exitCode" = 0 ]; then

				log "Backed up $folder"

				if [ "$backend" == "LOCAL" ]; then

					docker run --rm \
						--name restic \
						--hostname "$(hostname)" \
						-e XDG_CACHE_HOME=/cache \
						-e RESTIC_PASSWORD="$PASSWORD" \
						-v "$cachePath":/cache \
						-v "$backupPath":/repo \
						restic/restic check -r /repo

					exitCode=$?
				
				else

					docker run --rm \
						--name restic \
						--hostname "$(hostname)" \
						-e XDG_CACHE_HOME=/cache \
						-e RESTIC_PASSWORD="$PASSWORD" \
						-e "$remoteKeyName"="$REMOTE_KEY" \
						-e "$remoteSecretName"="$REMOTE_SECRET" \
						-v "$cachePath":/cache \
						restic/restic check -r "$remotePath"

					exitCode=$?
				
				fi

				if [ $exitCode != 0 ]; then
					errors=true
					log "Scan found errors in repo $backupPath"
				fi

				if [ "$backend" == "LOCAL" ]; then

					docker run --rm \
						--name restic \
						--hostname "$(hostname)" \
						-e XDG_CACHE_HOME=/cache \
						-e RESTIC_PASSWORD="$PASSWORD" \
						-v "$cachePath":/cache \
						-v "$backupPath":/repo \
						restic/restic forget --keep-daily 7 --keep-weekly 4 --keep-monthly 6 -r /repo

					exitCode=$?
				
				else

					docker run --rm \
						--name restic \
						--hostname "$(hostname)" \
						-e XDG_CACHE_HOME=/cache \
						-e RESTIC_PASSWORD="$PASSWORD" \
						-e "$remoteKeyName"="$REMOTE_KEY" \
						-e "$remoteSecretName"="$REMOTE_SECRET" \
						-v "$cachePath":/cache \
						restic/restic forget --keep-daily 7 --keep-weekly 4 --keep-monthly 6 -r "$remotePath"

					exitCode=$?

				fi

				if [ $exitCode != 0 ]; then
					errors=true
					log "Errors while pruning $backupPath"
				fi

			else

				errors=true
				log "Error: Failed to back up $folder"

			fi

		fi

	fi

done \
	< <(

		find "$PROJECTS_HOME" -name $COMPOSE_FILE_NAME -print0
	)

if $errors; then
	log "**** Finished backups with errors ****"
else
	log "**** Finished backups ****"
fi
